# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple

from nbxmpp.const import MessageType
from nbxmpp.errors import is_error
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import PubSubStanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.modules.util import finalize
from nbxmpp.modules.util import process_response
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Node
from nbxmpp.structs import CommonResult
from nbxmpp.structs import PubSubEventData
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task


class PubSub(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_pubsub_base,
                          ns=Namespace.PUBSUB_EVENT,
                          priority=15),
        ]

    def _process_pubsub_base(self, _client, stanza, properties):
        if properties.type not in (MessageType.HEADLINE, MessageType.NORMAL):
            return

        properties.pubsub = True
        event = stanza.getTag('event', namespace=Namespace.PUBSUB_EVENT)

        delete = event.getTag('delete')
        if delete is not None:
            node = delete.getAttr('node')
            properties.pubsub_event = PubSubEventData(
                node, deleted=True)
            return

        purge = event.getTag('purge')
        if purge is not None:
            node = purge.getAttr('node')
            properties.pubsub_event = PubSubEventData(node, purged=True)
            return

        items = event.getTag('items')
        if items is not None:
            node = items.getAttr('node')

            retract = items.getTag('retract')
            if retract is not None:
                id_ = retract.getAttr('id')
                properties.pubsub_event = PubSubEventData(
                    node, id_, retracted=True)
                return

            if len(items.getChildren()) != 1:
                self._log.warning('PubSub event with != 1 item')
                self._log.warning(stanza)
                return

            item = items.getTag('item')
            if item is None:
                self._log.warning('No item node found')
                self._log.warning(stanza)
                return
            id_ = item.getAttr('id')
            properties.pubsub_event = PubSubEventData(node, id_, item)

    @iq_request_task
    def request_item(self, node, id_, jid=None):
        task = yield

        response = yield _make_pubsub_request(node, id_=id_, jid=jid)

        if response.isError():
            raise PubSubStanzaError(response)

        item = _get_pubsub_item(response, node, id_)
        yield task.set_result(item)

    @iq_request_task
    def request_items(self, node, max_items=None, jid=None):
        _task = yield

        response = yield _make_pubsub_request(node,
                                              max_items=max_items,
                                              jid=jid)

        if response.isError():
            raise PubSubStanzaError(response)

        yield _get_pubsub_items(response, node)

    @iq_request_task
    def publish(self,
                node,
                item,
                id_=None,
                options=None,
                jid=None,
                force_node_options=False):

        _task = yield

        request = _make_publish_request(node, item, id_, options, jid)
        response = yield request

        if response.isError():
            error = PubSubStanzaError(response)
            if (not force_node_options or
                    error.app_condition != 'precondition-not-met'):
                raise error

            result = yield self.reconfigure_node(node, options, jid)
            if is_error(result):
                raise result

            response = yield request
            if response.isError():
                raise PubSubStanzaError(response)

        jid = response.getFrom()
        item_id = _get_published_item_id(response, node, id_)
        yield PubSubPublishResult(jid, node, item_id)

    @iq_request_task
    def get_access_model(self, node):
        _task = yield

        result = yield self.get_node_configuration(node)

        raise_if_error(result)

        yield result.form['pubsub#access_model'].value

    @iq_request_task
    def set_access_model(self, node, model):
        task = yield

        if model not in ('open', 'presence'):
            raise ValueError('Invalid access model')

        result = yield self.get_node_configuration(node)

        raise_if_error(result)

        try:
            access_model = result.form['pubsub#access_model'].value
        except Exception:
            yield task.set_error('warning',
                                 condition='access-model-not-supported')

        if access_model == model:
            jid = self._client.get_bound_jid().new_as_bare()
            yield CommonResult(jid=jid)

        result.form['pubsub#access_model'].value = model

        result = yield self.set_node_configuration(node, result.form)

        yield finalize(task, result)

    @iq_request_task
    def retract(self, node, id_, jid=None, notify=True):
        _task = yield

        response = yield _make_retract_request(node, id_, jid, notify)
        yield process_response(response)

    @iq_request_task
    def purge(self, node, jid=None):
        _task = yield

        response = yield _make_purge_request(node, jid)
        yield process_response(response)

    @iq_request_task
    def delete(self, node, jid=None):
        _task = yield

        response = yield _make_delete_request(node, jid)
        yield process_response(response)

    @iq_request_task
    def reconfigure_node(self, node, options, jid=None):
        _task = yield

        result = yield self.get_node_configuration(node, jid)
        if is_error(result):
            raise result

        _apply_options(result.form, options)
        result = yield self.set_node_configuration(node, result.form, jid)
        yield result

    @iq_request_task
    def set_node_configuration(self, node, form, jid=None):
        _task = yield

        response = yield _make_node_configuration(node, form, jid)
        yield process_response(response)

    @iq_request_task
    def get_node_configuration(self, node, jid=None):
        _task = yield

        response = yield _make_node_configuration_request(node, jid)

        if response.isError():
            raise PubSubStanzaError(response)

        jid = response.getFrom()
        form = _get_configure_form(response, node)
        yield PubSubNodeConfigurationResult(jid=jid, node=node, form=form)


def get_pubsub_request(jid, node, id_=None, max_items=None):
    query = Iq('get', to=jid)
    pubsub = query.addChild('pubsub', namespace=Namespace.PUBSUB)
    items = pubsub.addChild('items', {'node': node})
    if max_items is not None:
        items.setAttr('max_items', max_items)
    if id_ is not None:
        items.addChild('item', {'id': id_})
    return query


def get_pubsub_item(stanza):
    pubsub_node = stanza.getTag('pubsub')
    items_node = pubsub_node.getTag('items')
    return items_node.getTag('item')


def get_pubsub_items(stanza, node=None):
    pubsub_node = stanza.getTag('pubsub')
    items_node = pubsub_node.getTag('items')
    if node is not None and items_node.getAttr('node') != node:
        return None

    if items_node is not None:
        return items_node.getTags('item')
    return None


def get_publish_options(config):
    options = Node(Namespace.DATA + ' x', attrs={'type': 'submit'})
    field = options.addChild('field',
                             attrs={'var': 'FORM_TYPE', 'type': 'hidden'})
    field.setTagData('value', Namespace.PUBSUB_PUBLISH_OPTIONS)

    for var, value in config.items():
        field = options.addChild('field', attrs={'var': var})
        field.setTagData('value', value)
    return options


def _get_pubsub_items(response, node):
    pubsub_node = response.getTag('pubsub', namespace=Namespace.PUBSUB)
    if pubsub_node is None:
        raise MalformedStanzaError('pubsub node missing', response)

    items_node = pubsub_node.getTag('items')
    if items_node is None:
        raise MalformedStanzaError('items node missing', response)

    if items_node.getAttr('node') != node:
        raise MalformedStanzaError('invalid node attr', response)

    return items_node.getTags('item')


def _get_pubsub_item(response, node, id_):
    items = _get_pubsub_items(response, node)

    if len(items) > 1:
        raise MalformedStanzaError('multiple items found', response)

    if not items:
        return None

    item = items[0]
    if item.getAttr('id') != id_:
        raise MalformedStanzaError('invalid item id', response)

    return item


def _make_pubsub_request(node, id_=None, max_items=None, jid=None):
    query = Iq('get', to=jid)
    pubsub = query.addChild('pubsub', namespace=Namespace.PUBSUB)
    items = pubsub.addChild('items', {'node': node})
    if max_items is not None:
        items.setAttr('max_items', max_items)
    if id_ is not None:
        items.addChild('item', {'id': id_})
    return query


def _get_configure_form(response, node):
    pubsub = response.getTag('pubsub', namespace=Namespace.PUBSUB_OWNER)
    if pubsub is None:
        raise MalformedStanzaError('pubsub node missing', response)

    configure = pubsub.getTag('configure')
    if configure is None:
        raise MalformedStanzaError('configure node missing', response)

    if node != configure.getAttr('node'):
        raise MalformedStanzaError('invalid node attribute', response)

    forms = configure.getTags('x', namespace=Namespace.DATA)
    for form in forms:
        dataform = extend_form(node=form)
        form_type = dataform.vars.get('FORM_TYPE')
        if form_type is None or form_type.value != Namespace.PUBSUB_CONFIG:
            continue

        return dataform

    raise MalformedStanzaError('no valid form type found', response)


def _get_published_item_id(response, node, id_):
    pubsub = response.getTag('pubsub', namespace=Namespace.PUBSUB)
    if pubsub is None:
        # https://xmpp.org/extensions/xep-0060.html#publisher-publish-success
        # If the publish request did not include an ItemID,
        # the IQ-result SHOULD include an empty <item/> element
        # that specifies the ItemID of the published item.
        #
        # If the server did not add a payload we assume the item was
        # published with the id we requested
        return id_

    publish = pubsub.getTag('publish')
    if publish is None:
        raise MalformedStanzaError('publish node missing', response)

    if node != publish.getAttr('node'):
        raise MalformedStanzaError('invalid node attribute', response)

    item = publish.getTag('item')
    if item is None:
        raise MalformedStanzaError('item node missing', response)

    item_id = item.getAttr('id')
    if id_ is not None and item_id != id_:
        raise MalformedStanzaError('invalid item id', response)

    return item_id


def _make_publish_request(node, item, id_, options, jid):
    query = Iq('set', to=jid)
    pubsub = query.addChild('pubsub', namespace=Namespace.PUBSUB)
    publish = pubsub.addChild('publish', {'node': node})
    attrs = {}
    if id_ is not None:
        attrs = {'id': id_}
    publish.addChild('item', attrs, [item])
    if options:
        publish = pubsub.addChild('publish-options')
        publish.addChild(node=_make_publish_options(options))
    return query


def _make_publish_options(options):
    data = Node(Namespace.DATA + ' x', attrs={'type': 'submit'})
    field = data.addChild('field', attrs={'var': 'FORM_TYPE', 'type': 'hidden'})
    field.setTagData('value', Namespace.PUBSUB_PUBLISH_OPTIONS)

    for var, value in options.items():
        field = data.addChild('field', attrs={'var': var})
        field.setTagData('value', value)
    return data


def _make_retract_request(node, id_, jid, notify):
    query = Iq('set', to=jid)
    pubsub = query.addChild('pubsub', namespace=Namespace.PUBSUB)
    attrs = {'node': node}
    if notify:
        attrs['notify'] = 'true'
    retract = pubsub.addChild('retract', attrs=attrs)
    retract.addChild('item', {'id': id_})
    return query


def _make_purge_request(node, jid):
    query = Iq('set', to=jid)
    pubsub = query.addChild('pubsub', namespace=Namespace.PUBSUB_OWNER)

    pubsub.addChild('purge', attrs={'node': node})
    return query


def _make_delete_request(node, jid):
    query = Iq('set', to=jid)
    pubsub = query.addChild('pubsub', namespace=Namespace.PUBSUB_OWNER)

    pubsub.addChild('delete', attrs={'node': node})
    return query


def _make_node_configuration(node, form, jid):
    query = Iq('set', to=jid)
    pubsub = query.addChild('pubsub', namespace=Namespace.PUBSUB_OWNER)
    configure = pubsub.addChild('configure', {'node': node})
    form.setAttr('type', 'submit')
    configure.addChild(node=form)
    return query


def _make_node_configuration_request(node, jid):
    query = Iq('get', to=jid)
    pubsub = query.addChild('pubsub', namespace=Namespace.PUBSUB_OWNER)
    pubsub.addChild('configure', {'node': node})
    return query


def _apply_options(form, options):
    for var, value in options.items():
        try:
            field = form[var]
        except KeyError:
            pass
        else:
            field.value = value


PubSubNodeConfigurationResult = namedtuple('PubSubConfigResult',
                                           'jid node form')

PubSubConfigResult = namedtuple('PubSubConfigResult',
                                'jid node form')

PubSubPublishResult = namedtuple('PubSubPublishResult',
                                 'jid node id')



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import EMEData
from nbxmpp.structs import StanzaHandler


class EME(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_eme,
                          ns=Namespace.EME,
                          priority=40)
        ]

    def _process_eme(self, _client, stanza, properties):
        encryption = stanza.getTag('encryption', namespace=Namespace.EME)
        if encryption is None:
            return

        name = encryption.getAttr('name')
        namespace = encryption.getAttr('namespace')
        if namespace is None:
            self._log.warning('No namespace on message')
            return

        properties.eme = EMEData(name=name, namespace=namespace)
        self._log.info('Found data: %s', properties.eme)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.const import MessageType
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.fallback import parse_fallback_indication
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import BodyData
from nbxmpp.structs import StanzaHandler
from nbxmpp.structs import StanzaIDData
from nbxmpp.structs import XHTMLData
from nbxmpp.util import error_factory


class BaseMessage(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_message_base,
                          priority=5),
            StanzaHandler(name='message',
                          callback=self._process_message_after_base,
                          priority=10),
        ]

    def _process_message_base(self, _client, stanza, properties):
        properties.type = self._parse_type(stanza)

        if properties.is_carbon_message and properties.carbon.is_sent:
            properties.jid = stanza.getTo()

        elif properties.is_mam_message and not properties.type.is_groupchat:
            own_jid = self._client.get_bound_jid()
            if own_jid.bare_match(stanza.getFrom()):
                properties.jid = stanza.getTo()
            else:
                properties.jid = stanza.getFrom()

        else:
            properties.jid = stanza.getFrom()

        self._parse_if_private_message(stanza, properties)

        properties.remote_jid = self._determine_remote_jid(properties)
        properties.from_ = stanza.getFrom()
        properties.to = stanza.getTo()
        properties.id = stanza.getID()
        properties.self_message = self._parse_self_message(stanza, properties)

        properties.origin_id = stanza.getOriginID()
        properties.stanza_ids = self._parse_stanza_ids(stanza)

        if properties.type.is_error:
            properties.error = error_factory(stanza)

    def _determine_remote_jid(self, properties):
        if properties.is_muc_pm:
            return properties.jid
        return properties.jid.new_as_bare()

    def _parse_if_private_message(self, stanza, properties) -> None:
        muc_user = stanza.getTag('x', namespace=Namespace.MUC_USER)
        if muc_user is None:
            return

        if not properties.jid.is_full:
            return

        if (properties.type.is_chat or
                properties.type.is_error and
                not muc_user.getChildren()):
            properties.muc_private_message = True

    def _process_message_after_base(self, _client, stanza: Message, properties):
        # This handler runs after decryption handlers had the chance
        # to decrypt the body

        fallbacks_for = parse_fallback_indication(self._log, stanza)

        properties.body = stanza.getBody()
        properties.bodies = BodyData(
            stanza, fallbacks_for, self._client.get_supported_fallback_ns())
        properties.thread = stanza.getThread()
        properties.subject = stanza.getSubject()
        forms = stanza.getTags('x', namespace=Namespace.DATA)
        if forms:
            properties.forms = forms

        xhtml = stanza.getXHTML()
        if xhtml is None:
            return

        if xhtml.getTag('body', namespace=Namespace.XHTML) is None:
            self._log.warning('xhtml without body found')
            self._log.warning(stanza)
            return

        properties.xhtml = XHTMLData(xhtml)

    def _parse_type(self, stanza):
        type_ = stanza.getType()
        if type_ is None:
            return MessageType.NORMAL

        try:
            return MessageType(type_)
        except ValueError:
            self._log.warning('Message with invalid type: %s', type_)
            self._log.warning(stanza)
            raise NodeProcessed

    @staticmethod
    def _parse_self_message(stanza, properties):
        if properties.type.is_groupchat:
            return False
        return stanza.getFrom().bare_match(stanza.getTo())

    def _parse_stanza_ids(self, stanza):
        stanza_ids = []
        for stanza_id in stanza.getTags('stanza-id', namespace=Namespace.SID):
            id_ = stanza_id.getAttr('id')
            by = stanza_id.getAttr('by')
            if not id_ or not by:
                self._log.warning('Missing attributes on stanza-id')
                self._log.warning(stanza)
                continue

            stanza_ids.append(StanzaIDData(id=id_, by=by))

        return stanza_ids



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.const import AdHocAction
from nbxmpp.const import AdHocNoteType
from nbxmpp.const import AdHocStatus
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.discovery import get_disco_request
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Node
from nbxmpp.structs import AdHocCommand
from nbxmpp.structs import AdHocCommandNote
from nbxmpp.task import iq_request_task


class AdHoc(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_command_list(self, jid=None):
        _task = yield

        if jid is None:
            jid = self._client.get_bound_jid().bare
        response = yield get_disco_request(Namespace.DISCO_ITEMS,
                                           jid,
                                           node=Namespace.COMMANDS)
        if response.isError():
            raise StanzaError(response)

        children = response.getQueryChildren()

        command_list = []
        for item in children:
            if item.getName() != 'item':
                continue
            try:
                command_list.append(AdHocCommand(**item.getAttrs()))
            except Exception as error:
                raise MalformedStanzaError(f'invalid item attributes: {error}',
                                           response)

        yield command_list

    @iq_request_task
    def execute_command(self, cmd, action=None, dataform=None):
        _task = yield

        if action is None:
            action = AdHocAction.EXECUTE
        attrs = {'node': cmd.node,
                 'xmlns': Namespace.COMMANDS,
                 'action': action.value}
        if cmd.sessionid is not None:
            attrs['sessionid'] = cmd.sessionid

        response = yield _make_command(cmd, attrs, dataform)
        if response.isError():
            raise StanzaError(response)

        command = response.getTag('command', namespace=Namespace.COMMANDS)
        if command is None:
            raise MalformedStanzaError('command node missing', response)

        node = command.getAttr('node')
        if node is None:
            raise MalformedStanzaError('node attribute missing', response)

        status = command.getAttr('status')
        if status is None:
            raise MalformedStanzaError('status attribute missing', response)

        if status not in ('executing', 'completed', 'canceled'):
            raise MalformedStanzaError('invalid status attribute %s' % status,
                                       response)

        status = AdHocStatus(status)

        sessionid = command.getAttr('sessionid')
        if sessionid is None and _expect_sessionid(status, cmd.sessionid):
            raise MalformedStanzaError('sessionid attribute missing', response)

        try:
            notes = _parse_notes(command)
        except ValueError as error:
            raise MalformedStanzaError(error, response)

        try:
            actions, default = _parse_actions(command)
        except ValueError as error:
            raise MalformedStanzaError(error, response)

        yield AdHocCommand(
            jid=response.getFrom(),
            name=None,
            node=node,
            sessionid=sessionid,
            status=status,
            data=command.getTag('x', namespace=Namespace.DATA),
            actions=actions,
            default=default,
            notes=notes)


def _make_command(command, attrs, dataform):
    command_node = Node('command', attrs=attrs)
    if dataform is not None:
        command_node.addChild(node=dataform)
    iq = Iq('set', to=command.jid)
    iq.addChild(node=command_node)
    return iq


def _parse_notes(command):
    notes = []
    for note in command.getTags('note'):
        type_ = note.getAttr('type')
        if type_ is None:
            type_ = 'info'

        if type_ not in ('info', 'warn', 'error'):
            raise ValueError('invalid note type %s' % type_)

        notes.append(AdHocCommandNote(text=note.getData(),
                                      type=AdHocNoteType(type_)))
    return notes


def _parse_actions(command):
    if command.getAttr('status') != 'executing':
        return set(), None

    actions_node = command.getTag('actions')
    if actions_node is None:
        # If there is no <actions/> element,
        # the user-agent can use a single-stage dialog or view.
        # The action "execute" is equivalent to the action "complete".
        return {AdHocAction.CANCEL, AdHocAction.COMPLETE}, AdHocAction.COMPLETE

    default = actions_node.getAttr('execute')
    if default is None:
        # If the "execute" attribute is absent, it defaults to "next".
        default = 'next'

    if default not in ('prev', 'next', 'complete'):
        raise ValueError('invalid execute attribute %s' % default)

    default = AdHocAction(default)

    # We use a set because it cannot contain duplicates
    actions = set()
    for action in actions_node.getChildren():
        name = action.getName()
        if name == 'execute':
            actions.add(default)

        if name in ('prev', 'next', 'complete'):
            actions.add(AdHocAction(name))

    if not actions:
        raise ValueError('actions element without actions')

    # The action "cancel" is always allowed.
    actions.add(AdHocAction.CANCEL)

    # A form which has an <actions/> element and an "execute" attribute
    # which evaluates (taking the default into account if absent) to an
    # action which is not allowed is therefore invalid.
    if default not in actions:
        # Some implementations don’t respect this rule.
        # Take the first action so we don’t fail here.
        for act in actions:
            default = act
            break

    return actions, default


def _expect_sessionid(status, sent_sessionid):
    # Session id should only be expected for multiple stage commands
    # or when we initialize the session (set the session attribute)
    return status != status.COMPLETED or sent_sessionid is not None



from __future__ import annotations

from test.lib.util import StanzaHandlerTest

import nbxmpp
from nbxmpp.language import LanguageRange
from nbxmpp.structs import BodyData


class TestBody(StanzaHandlerTest):

    def test_body_language(self):
        text = '> Anna wrote:\n> Hi, how are you?\nGreat'
        text_de = '> Anna schrieb\n>Hallo, wie geht es dir?\nGut'

        xml = '''
            <message to='anna@example.com' id='message-id2' type='groupchat'>
              <body>%s</body>
              <body xml:lang='en-us'>%s</body>
              <body xml:lang='de-DE'>%s</body>
            </message>
        ''' % (text, text, text_de)

        message = nbxmpp.Message(node=xml)

        body_data = BodyData(message, None, None)
        body = body_data.get([LanguageRange(tag='en')])
        self.assertEqual(body, text)

        body = body_data.get(None)
        self.assertEqual(body, text)

        body = body_data.get([LanguageRange(tag='de')])
        self.assertEqual(body, text_de)

        body = body_data.get([LanguageRange(tag='ar')])
        self.assertEqual(body, text)

        xml = '''
            <message to='anna@example.com' id='message-id2' type='groupchat'>
              <body xml:lang='en-us'>en</body>
              <body xml:lang='de-DE'>de</body>
            </message>
        '''

        message = nbxmpp.Message(node=xml)

        body_data = BodyData(message)
        body = body_data.get([LanguageRange(tag='en')])
        self.assertEqual(body, 'en')

        # Language not in body, pick any language
        body_data = BodyData(message)
        body = body_data.get([LanguageRange(tag='ar')])
        self.assertEqual(body, 'de')

        # Supply range with priority
        body_data = BodyData(message)
        body = body_data.get([LanguageRange(tag='de'), LanguageRange(tag='en')])
        self.assertEqual(body, 'de')

        body_data = BodyData(message)
        body = body_data.get([LanguageRange(tag='ar'), LanguageRange(tag='en')])
        self.assertEqual(body, 'en')



from nbxmpp import Namespace
from nbxmpp import Node
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import RetractionData
from nbxmpp.structs import StanzaHandler


class Retraction(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name='message',
                callback=self._process_message,
                ns=Namespace.MESSAGE_RETRACT_1,
                priority=20,
            ),
            StanzaHandler(
                name='message',
                callback=self._process_message_retracted_tombstone,
                ns=Namespace.MESSAGE_RETRACT_1,
                priority=20,
            ),
        ]

    def _process_message(
        self, _client, stanza: Node, properties: MessageProperties
    ) -> None:
        retraction = stanza.getTag(
            'retract', namespace=Namespace.MESSAGE_RETRACT_1
        )

        if retraction is None:
            return

        retracted_id = retraction.getAttr('id')

        if retracted_id is None:
            self._log.warning('<retract> without retracted message id')
            return

        properties.retraction = RetractionData(
            id=retracted_id, is_tombstone=False, timestamp=None
        )

    def _process_message_retracted_tombstone(
        self, _client, stanza: Node, properties: MessageProperties
    ) -> None:
        if not properties.is_mam_message:
            return

        retracted = stanza.getTag(
            'retracted', namespace=Namespace.MESSAGE_RETRACT_1
        )

        if retracted is None:
            return

        retracted_stamp = retracted.getAttr('stamp')

        properties.retraction = RetractionData(
            id=None,
            is_tombstone=True,
            timestamp=parse_datetime(
                retracted_stamp, check_utc=True, convert='utc', epoch=True
            ),
        )



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import Node
from nbxmpp.structs import AnnotationNote
from nbxmpp.task import iq_request_task


class Annotations(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @property
    def domain(self):
        return self._client.get_bound_jid().domain

    @iq_request_task
    def request_annotations(self):
        _task = yield

        response = yield _make_request()
        if response.isError():
            raise StanzaError(response)

        query = response.getQuery()
        storage = query.getTag('storage', namespace=Namespace.ROSTERNOTES)
        if storage is None:
            raise MalformedStanzaError('storage node missing', response)

        notes = []
        for note in storage.getTags('note'):
            try:
                jid = JID.from_string(note.getAttr('jid'))
            except Exception as error:
                self._log.warning('Invalid JID: %s, %s',
                                  note.getAttr('jid'), error)
                continue

            cdate = note.getAttr('cdate')
            if cdate is not None:
                cdate = parse_datetime(cdate, epoch=True)

            mdate = note.getAttr('mdate')
            if mdate is not None:
                mdate = parse_datetime(mdate, epoch=True)

            data = note.getData()
            notes.append(AnnotationNote(jid=jid, cdate=cdate,
                                        mdate=mdate, data=data))

        self._log.info('Received annotations from %s:', self.domain)
        for note in notes:
            self._log.info(note)
        yield notes

    @iq_request_task
    def set_annotations(self, notes):
        _task = yield

        for note in notes:
            self._log.info(note)

        response = yield _make_set_request(notes)
        yield process_response(response)


def _make_request():
    payload = Node('storage', attrs={'xmlns': Namespace.ROSTERNOTES})
    return Iq(typ='get', queryNS=Namespace.PRIVATE, payload=payload)

def _make_set_request(notes):
    storage = Node('storage', attrs={'xmlns': Namespace.ROSTERNOTES})
    for note in notes:
        node = Node('note', attrs={'jid': note.jid})
        node.setData(note.data)
        if note.cdate is not None:
            node.setAttr('cdate', note.cdate)
        if note.mdate is not None:
            node.setAttr('mdate', note.mdate)
        storage.addChild(node=node)
    return Iq(typ='set', queryNS=Namespace.PRIVATE, payload=storage)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

import logging

from nbxmpp.modules.delay import parse_delay
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import InvalidFrom
from nbxmpp.protocol import InvalidStanza
from nbxmpp.protocol import Message
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import CarbonData
from nbxmpp.structs import MAMData

log = logging.getLogger('nbxmpp.m.misc')


def unwrap_carbon(stanza, own_jid):
    carbon = stanza.getTag('received', namespace=Namespace.CARBONS)
    if carbon is None:
        carbon = stanza.getTag('sent', namespace=Namespace.CARBONS)
        if carbon is None:
            return stanza, None

    # Carbon must be from our bare jid
    if stanza.getFrom() != own_jid.new_as_bare():
        raise InvalidFrom('Invalid from: %s' % stanza.getAttr('from'))

    forwarded = carbon.getTag('forwarded', namespace=Namespace.FORWARD)
    message = Message(node=forwarded.getTag('message'))

    type_ = carbon.getName()

    # Fill missing to/from
    to = message.getTo()
    if to is None:
        message.setTo(own_jid.bare)

    frm = message.getFrom()
    if frm is None:
        message.setFrom(own_jid.bare)

    if type_ == 'received':
        if message.getFrom().bare_match(own_jid):
            # Drop 'received' Carbons from ourself, we already
            # got the message with the 'sent' Carbon or via the
            # message itself
            raise NodeProcessed('Drop "received"-Carbon from ourself')

        if message.getTag('x', namespace=Namespace.MUC_USER) is not None:
            # A MUC broadcasts messages sent to us to all resources
            # there is no need to process the received carbon
            raise NodeProcessed('Drop MUC-PM "received"-Carbon')

    return message, CarbonData(type=type_)


def unwrap_mam(stanza, own_jid):
    result = stanza.getTag('result', namespace=Namespace.MAM_2)
    if result is None:
        result = stanza.getTag('result', namespace=Namespace.MAM_1)
        if result is None:
            return stanza, None

    query_id = result.getAttr('queryid')
    if query_id is None:
        log.warning('No queryid on MAM message')
        log.warning(stanza)
        raise InvalidStanza

    id_ = result.getAttr('id')
    if id_ is None:
        log.warning('No id on MAM message')
        log.warning(stanza)
        raise InvalidStanza

    forwarded = result.getTag('forwarded', namespace=Namespace.FORWARD)
    message = Message(node=forwarded.getTag('message'))

    # Fill missing to/from
    to = message.getTo()
    if to is None:
        message.setTo(own_jid.bare)

    frm = message.getFrom()
    if frm is None:
        message.setFrom(own_jid.bare)

    # Timestamp parsing
    # Most servers dont set the 'from' attr, so we cant check for it
    delay_timestamp = parse_delay(forwarded, epoch=True)
    if delay_timestamp is None:
        log.warning('No timestamp on MAM message')
        log.warning(stanza)
        raise InvalidStanza

    return message, MAMData(id=id_,
                            query_id=query_id,
                            archive=stanza.getFrom(),
                            namespace=result.getNamespace(),
                            timestamp=delay_timestamp)


def build_xhtml_body(xhtml, xmllang=None):
    try:
        if xmllang is not None:
            body = '<body xmlns="%s" xml:lang="%s">%s</body>' % (
                Namespace.XHTML, xmllang, xhtml)
        else:
            body = '<body xmlns="%s">%s</body>' % (Namespace.XHTML, xhtml)
    except Exception as error:
        log.error('Error while building xhtml node: %s', error)
        return None
    return body



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

import logging
from dataclasses import dataclass
from dataclasses import field

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.simplexml import Node
from nbxmpp.task import iq_request_task

log = logging.getLogger('nbxmpp.m.vcard4')


ALLOWED_SEX_VALUES = ['M', 'F', 'O', 'N', 'U']
ALLOWED_KIND_VALUES = ['individual', 'group', 'org', 'location']

# Cardinality
# 1     Exactly one instance per vCard MUST be present.
# *1    Exactly one instance per vCard MAY be present.
# 1*    One or more instances per vCard MUST be present.
# *     One or more instances per vCard MAY be present.

PROPERTY_DEFINITION = {
    'source': (['altid', 'pid', 'pref', 'mediatype'], '*'),
    'kind': ([], '*1'),
    'fn': (['language', 'altid', 'pid', 'pref', 'type'], '1*'),
    'n': (['language', 'altid', 'sort-as'], '*1'),
    'nickname': (['language', 'altid', 'pid', 'pref', 'type'], '*'),
    'photo': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'bday': (['altid', 'calscale'], '*1'),
    'anniversary': (['altid', 'calscale'], '*1'),
    'gender': ([], '*1'),
    'adr': (['language', 'altid', 'pid', 'pref', 'type', 'geo', 'tz', 'label'], '*'),
    'tel': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'email': (['altid', 'pid', 'pref', 'type'], '*'),
    'impp': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'lang': (['altid', 'pid', 'pref', 'type'], '*'),
    'tz': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'geo': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'title': (['language', 'altid', 'pid', 'pref', 'type'], '*'),
    'role': (['language', 'altid', 'pid', 'pref', 'type'], '*'),
    'logo': (['language', 'altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'org': (['language', 'altid', 'pid', 'pref', 'type', 'sort-as'], '*'),
    'member': (['altid', 'pid', 'pref', 'mediatype'], '*'),
    'related': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'categories': (['altid', 'pid', 'pref', 'type'], '*'),
    'note': (['language', 'altid', 'pid', 'pref', 'type'], '*'),
    'prodid': ([], '*1'),
    'rev': ([], '*1'),
    'sound': (['language', 'altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'uid': ([], '*1'),
    'clientpidmap': ([], '*'),
    'url': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'key': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'fburl': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'caladruri': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
    'caluri': (['altid', 'pid', 'pref', 'type', 'mediatype'], '*'),
}


PROPERTY_VALUE_TYPES = {
    'bday': ['date', 'time', 'date-time', 'text'],
    'anniversary': ['date', 'time', 'date-time', 'text'],
    'key': ['text', 'uri'],
    'tel': ['uri', 'text'],
    'tz': ['text', 'uri', 'utc-offset'],
    'related': ['text', 'uri'],
}


def get_data_from_children(node, child_name):
    values = []
    child_nodes = node.getTags(child_name)
    for child_node in child_nodes:
        child_value = child_node.getData()
        if child_value:
            values.append(child_value)
    return values


def add_children(node, child_name, values):
    for value in values:
        node.addChild(child_name, payload=value)


def get_multiple_type_value(node, types):
    for type_ in types:
        value = node.getTagData(type_)
        if value:
            return type_, value

    raise ValueError('no value found')


def make_parameters(parameters):
    parameters_node = Node('parameters')
    for parameter in parameters.values():
        parameters_node.addChild(node=parameter.to_node())
    return parameters_node


def get_parameters(node):
    name = node.getName()
    definition = PROPERTY_DEFINITION[name]
    allowed_parameters = definition[0]
    parameters_node = node.getTag('parameters')
    if parameters_node is None:
        return Parameters()

    parameters = {}
    for parameter in allowed_parameters:
        parameter_node = parameters_node.getTag(parameter)
        if parameter_node is None:
            continue

        parameter_class = PARAMETER_CLASSES.get(parameter_node.getName())
        if parameter_class is None:
            continue

        parameter = parameter_class.from_node(parameter_node)
        parameters[parameter.name] = parameter

    return Parameters(parameters)


@dataclass
class Parameter:

    name: str
    type: str
    value: str

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid parameter name: {name}')

        value = node.getTagData(cls.type)
        if not value:
            raise ValueError('no parameter value found')

        return cls(value)

    def to_node(self):
        node = Node(self.name)
        node.addChild(self.type, payload=self.value)
        return node

    def copy(self):
        return self.__class__(value=self.value)


@dataclass
class MultiParameter:

    name: str
    type: str
    values: set[str]

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid parameter name: {name}')

        value_nodes = node.getTags(cls.type)
        if not value_nodes:
            raise ValueError('no parameter value found')

        values = set()
        for value_node in value_nodes:
            value = value_node.getData()
            if value:
                values.add(value)

        if not values:
            raise ValueError('no parameter value found')

        return cls(values)

    def to_node(self):
        node = Node(self.name)
        for value in self.values:
            node.addChild(self.type, payload=value)
        return node

    def copy(self):
        return self.__class__(values=set(self.values))


@dataclass
class LanguageParameter(Parameter):

    name: str = field(default='language', init=False)
    type: str = field(default='language-tag', init=False)


@dataclass
class PrefParameter(Parameter):

    name: str = field(default='pref', init=False)
    type: str = field(default='integer', init=False)


@dataclass
class AltidParameter(Parameter):

    name: str = field(default='altid', init=False)
    type: str = field(default='text', init=False)


@dataclass
class PidParameter(MultiParameter):

    name: str = field(default='pid', init=False)
    type: str = field(default='text', init=False)


@dataclass
class TypeParameter(MultiParameter):

    name: str = field(default='type', init=False)
    type: str = field(default='text', init=False)


@dataclass
class MediatypeParameter(Parameter):

    name: str = field(default='mediatype', init=False)
    type: str = field(default='text', init=False)


@dataclass
class CalscaleParameter(Parameter):

    name: str = field(default='calscale', init=False)
    type: str = field(default='text', init=False)


@dataclass
class SortasParameter(MultiParameter):

    name: str = field(default='sort-as', init=False)
    type: str = field(default='text', init=False)


@dataclass
class GeoParameter(Parameter):

    name: str = field(default='geo', init=False)
    type: str = field(default='uri', init=False)


@dataclass
class TzParameter:

    name: str = field(default='tz', init=False)
    value_type: str
    value: str

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        value_type, value = get_multiple_type_value(node, ['text', 'uri'])
        return cls(value_type, value)

    def to_node(self):
        node = Node(self.name)
        node.addChild(self.value_type, payload=self.value)
        return node

    def copy(self):
        return self.__class__(value_type=self.value_type,
                              value=self.value)


class Parameters:
    def __init__(self, parameters=None):
        if parameters is None:
            parameters = {}
        self._parameters = parameters

    def values(self):
        return self._parameters.values()

    def get_types(self):
        parameter = self._parameters.get('type')
        if parameter is None:
            return set()

        return parameter.values

    def remove_types(self, types):
        parameter = self._parameters.get('type')
        if parameter is None:
            raise ValueError('no type parameter')

        for type_ in types:
            parameter.values.discard(type_)

        if not parameter.values:
            self._parameters.pop('type')

    def add_types(self, types):
        parameter = self._parameters.get('type')
        if parameter is None:
            parameter = TypeParameter(set(types))
            self._parameters['type'] = parameter
            return

        parameter.values.update(types)

    def copy(self):
        parameters = {}
        for name, parameter in self._parameters.items():
            parameters[name] = parameter.copy()
        return self.__class__(parameters=parameters)


PARAMETER_CLASSES = {
    'language': LanguageParameter,
    'pref': PrefParameter,
    'altid': AltidParameter,
    'pid': PidParameter,
    'type': TypeParameter,
    'mediatype': MediatypeParameter,
    'calscale': CalscaleParameter,
    'sort-as': SortasParameter,
    'geo': GeoParameter,
    'tz': TzParameter
}


@dataclass
class UriProperty:

    name: str
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        value = node.getTagData('uri')
        if not value:
            raise ValueError('no value found')

        parameters = get_parameters(node)

        return cls(value, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild('uri', payload=self.value)
        return node

    @property
    def is_empty(self):
        return not self.value

    def copy(self):
        return self.__class__(value=self.value,
                              parameters=self.parameters.copy())


@dataclass
class TextProperty:

    name: str
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        text = node.getTagData('text')
        if not text:
            raise ValueError('no value found')

        parameters = get_parameters(node)

        return cls(text, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild('text', payload=self.value)
        return node

    @property
    def is_empty(self):
        return not self.value

    def copy(self):
        return self.__class__(value=self.value,
                              parameters=self.parameters.copy())


@dataclass
class TextListProperty:

    name: str
    values: list[str]
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        text_nodes = node.getTags('text')
        if not text_nodes:
            raise ValueError('no value found')

        values = get_data_from_children(node, 'text')
        if not values:
            raise ValueError('no values found')

        parameters = get_parameters(node)

        return cls(values, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        add_children(node, 'text', self.values)
        return node

    @property
    def is_empty(self):
        return not self.values

    def copy(self):
        return self.__class__(values=list(self.values),
                              parameters=self.parameters.copy())


@dataclass
class MultipleValueProperty:

    name: str
    value_type: str
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        types = PROPERTY_VALUE_TYPES[cls.name]
        value_type, value = get_multiple_type_value(node, types)

        parameters = get_parameters(node)

        return cls(value_type, value, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild(self.value_type, payload=self.value)
        return node

    @property
    def is_empty(self):
        return not self.value

    def copy(self):
        return self.__class__(value_type=self.value_type,
                              value=self.value,
                              parameters=self.parameters.copy())


@dataclass
class SourceProperty(UriProperty):

    name: str = field(default='source', init=False)


@dataclass
class KindProperty(TextProperty):

    name: str = field(default='kind', init=False)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        text = node.getTagData('text')
        if not text:
            raise ValueError('no value found')

        if text not in ALLOWED_KIND_VALUES:
            text = 'individual'

        parameters = get_parameters(node)

        return cls(text, parameters)


@dataclass
class FnProperty(TextProperty):

    name: str = field(default='fn', init=False)


@dataclass
class NProperty:

    name: str = field(default='n', init=False)
    surname: list[str] = field(default_factory=list)
    given: list[str] = field(default_factory=list)
    additional: list[str] = field(default_factory=list)
    prefix: list[str] = field(default_factory=list)
    suffix: list[str] = field(default_factory=list)
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        surname = get_data_from_children(node, 'surname')
        given = get_data_from_children(node, 'given')
        additional = get_data_from_children(node, 'additional')
        prefix = get_data_from_children(node, 'prefix')
        suffix = get_data_from_children(node, 'suffix')

        parameters = get_parameters(node)

        return cls(surname, given, additional, prefix, suffix, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        add_children(node, 'surname', self.surname)
        add_children(node, 'given', self.given)
        add_children(node, 'additional', self.additional)
        add_children(node, 'prefix', self.prefix)
        add_children(node, 'suffix', self.suffix)
        return node

    @property
    def is_empty(self):
        if (self.surname or
                self.given or
                self.additional or
                self.suffix or
                self.prefix):
            return False
        return True

    def copy(self):
        return self.__class__(surname=list(self.surname),
                              given=list(self.given),
                              additional=list(self.additional),
                              prefix=list(self.prefix),
                              suffix=list(self.suffix),
                              parameters=self.parameters.copy())

@dataclass
class NicknameProperty(TextListProperty):

    name: str = field(default='nickname', init=False)


@dataclass
class PhotoProperty(UriProperty):

    name: str = field(default='photo', init=False)


@dataclass
class BDayProperty(MultipleValueProperty):

    name: str = field(default='bday', init=False)


@dataclass
class AnniversaryProperty(MultipleValueProperty):

    name: str = field(default='anniversary', init=False)


@dataclass
class GenderProperty:

    name: str = field(default='gender', init=False)
    sex: Optional[str] = None
    identity: Optional[str] = None
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        sex = node.getTagData('sex')
        if sex not in ALLOWED_SEX_VALUES:
            sex = None

        identity = node.getTagData('identity')
        if not identity:
            identity = None

        parameters = get_parameters(node)

        return cls(sex, identity, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        if self.sex:
            node.addChild('sex', payload=self.sex)
        if self.identity:
            node.addChild('identity', payload=self.identity)
        return node

    @property
    def is_empty(self):
        if self.sex or self.identity:
            return False
        return True

    def copy(self):
        return self.__class__(sex=self.sex,
                              identity=self.identity,
                              parameters=self.parameters.copy())


@dataclass
class AdrProperty:

    name: str = field(default='adr', init=False)
    pobox: list[str] = field(default_factory=list)
    ext: list[str] = field(default_factory=list)
    street: list[str] = field(default_factory=list)
    locality: list[str] = field(default_factory=list)
    region: list[str] = field(default_factory=list)
    code: list[str] = field(default_factory=list)
    country: list[str] = field(default_factory=list)
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        pobox = get_data_from_children(node, 'pobox')
        ext = get_data_from_children(node, 'ext')
        street = get_data_from_children(node, 'street')
        locality = get_data_from_children(node, 'locality')
        region = get_data_from_children(node, 'region')
        code = get_data_from_children(node, 'code')
        country = get_data_from_children(node, 'country')

        parameters = get_parameters(node)

        return cls(pobox, ext, street, locality,
                   region, code, country, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        add_children(node, 'pobox', self.pobox)
        add_children(node, 'ext', self.ext)
        add_children(node, 'street', self.street)
        add_children(node, 'locality', self.locality)
        add_children(node, 'region', self.region)
        add_children(node, 'code', self.code)
        add_children(node, 'country', self.country)
        return node

    @property
    def is_empty(self):
        if (self.pobox or
                self.ext or
                self.street or
                self.locality or
                self.region or
                self.code or
                self.country):
            return False
        return True

    def copy(self):
        return self.__class__(pobox=list(self.pobox),
                              ext=list(self.ext),
                              street=list(self.street),
                              locality=list(self.locality),
                              region=list(self.region),
                              code=list(self.code),
                              country=list(self.country),
                              parameters=self.parameters.copy())


@dataclass
class TelProperty(MultipleValueProperty):

    name: str = field(default='tel', init=False)


@dataclass
class EmailProperty(TextProperty):

    name: str = field(default='email', init=False)


@dataclass
class ImppProperty(UriProperty):

    name: str = field(default='impp', init=False)


@dataclass
class LangProperty:

    name: str = field(default='lang', init=False)
    value: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        value = node.getTagData('language-tag')
        if not value:
            raise ValueError('no value found')

        parameters = get_parameters(node)

        return cls(value, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild('language-tag', payload=self.value)
        return node

    @property
    def is_empty(self):
        return not self.value

    def copy(self):
        return self.__class__(value=self.value,
                              parameters=self.parameters.copy())


@dataclass
class TzProperty(MultipleValueProperty):

    name: str = field(default='tz', init=False)


@dataclass
class GeoProperty(UriProperty):

    name: str = field(default='geo', init=False)


@dataclass
class TitleProperty(TextProperty):

    name: str = field(default='title', init=False)


@dataclass
class RoleProperty(TextProperty):

    name: str = field(default='role', init=False)


@dataclass
class LogoProperty(UriProperty):

    name: str = field(default='logo', init=False)


@dataclass
class OrgProperty(TextListProperty):

    name: str = field(default='org', init=False)


@dataclass
class MemberProperty(UriProperty):

    name: str = field(default='member', init=False)


@dataclass
class RelatedProperty(MultipleValueProperty):

    name: str = field(default='related', init=False)


@dataclass
class CategoriesProperty(TextListProperty):

    name: str = field(default='categories', init=False)


@dataclass
class NoteProperty(TextProperty):

    name: str = field(default='note', init=False)


@dataclass
class ProdidProperty(TextProperty):

    name: str = field(default='prodid', init=False)


@dataclass
class RevProperty(TextProperty):

    name: str = field(default='rev', init=False)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        timestamp = node.getTagData('timestamp')
        if not timestamp:
            raise ValueError('no value found')

        parameters = get_parameters(node)

        return cls(timestamp, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild('timestamp', payload=self.value)
        return node


@dataclass
class SoundProperty(UriProperty):

    name: str = field(default='sound', init=False)


@dataclass
class UidProperty(UriProperty):

    name: str = field(default='uid', init=False)


@dataclass
class ClientpidmapProperty:

    name: str = field(default='clientpidmap', init=False)
    sourceid: int
    uri: str
    parameters: Parameters = field(default_factory=Parameters)

    @classmethod
    def from_node(cls, node):
        name = node.getName()
        if name != cls.name:
            raise ValueError(f'invalid property name: {name}')

        sourceid = node.getTagData('sourceid')
        if not sourceid:
            raise ValueError('no value found')

        uri = node.getTagData('uri')
        if not uri:
            raise ValueError('no value found')

        parameters = get_parameters(node)

        return cls(sourceid, uri, parameters)

    def to_node(self):
        node = Node(self.name)
        if self.parameters:
            node.addChild(node=make_parameters(self.parameters))
        node.addChild('sourceid', payload=self.sourceid)
        node.addChild('uri', payload=self.uri)
        return node

    @property
    def is_empty(self):
        return not self.uri

    def copy(self):
        return self.__class__(sourceid=self.sourceid,
                              uri=self.uri,
                              parameters=self.parameters.copy())


@dataclass
class UrlProperty(UriProperty):

    name: str = field(default='url', init=False)


@dataclass
class KeyProperty(MultipleValueProperty):

    name: str = field(default='key', init=False)


@dataclass
class FBurlProperty(UriProperty):

    name: str = field(default='fburl', init=False)


@dataclass
class CaladruriProperty(UriProperty):

    name: str = field(default='caladruri', init=False)


@dataclass
class CaluriProperty(UriProperty):

    name: str = field(default='calurl', init=False)


PROPERTY_CLASSES = {
    'source': SourceProperty,
    'kind': KindProperty,
    'fn': FnProperty,
    'n': NProperty,
    'nickname': NicknameProperty,
    'photo': PhotoProperty,
    'bday': BDayProperty,
    'anniversary': AnniversaryProperty,
    'gender': GenderProperty,
    'adr': AdrProperty,
    'tel': TelProperty,
    'email': EmailProperty,
    'impp': ImppProperty,
    'lang': LangProperty,
    'tz': TzProperty,
    'geo': GeoProperty,
    'title': TitleProperty,
    'role': RoleProperty,
    'logo': LogoProperty,
    'org': OrgProperty,
    'member': MemberProperty,
    'related': RelatedProperty,
    'categories': CategoriesProperty,
    'note': NoteProperty,
    'prodid': ProdidProperty,
    'rev': RevProperty,
    'sound': SoundProperty,
    'uid': UidProperty,
    'clientpidmap': ClientpidmapProperty,
    'url': UrlProperty,
    'key': KeyProperty,
    'fburl': FBurlProperty,
    'caladruri': CaladruriProperty,
    'caluri': CaluriProperty,
}


def get_property_from_name(name, node):
    property_class = PROPERTY_CLASSES.get(name)
    if property_class is None:
        return None

    try:
        return property_class.from_node(node)
    except Exception as error:
        log.warning('invalid vcard property: %s %s', error, node)
        return None


class VCard:
    def __init__(self, properties=None):
        if properties is None:
            properties = []
        self._properties = properties

    @classmethod
    def from_node(cls, node):
        properties = []
        for child in node.getChildren():
            child_name = child.getName()

            if child_name == 'group':
                group_name = child.getAttr('name')
                if not group_name:
                    continue

                group_properties = []
                for group_child in child.getChildren():
                    group_child_name = group_child.getName()
                    property_ = get_property_from_name(group_child_name,
                                                       group_child)
                    if property_ is None:
                        continue
                    group_properties.append(property_)

                properties.append((group_name, group_properties))

            else:

                property_ = get_property_from_name(child_name, child)
                if property_ is None:
                    continue
                properties.append((None, property_))

        return cls(properties)

    def to_node(self):
        vcard = Node(f'{Namespace.VCARD4} vcard')
        for group, props in self._properties:
            if group is None:
                vcard.addChild(node=props.to_node())
            else:
                group = Node(group)
                for prop in props:
                    group.addChild(node=prop.to_node())
                vcard.addChild(node=group)
        return vcard

    def get_properties(self):
        properties = []
        for group, props in self._properties:
            if group is None:
                properties.append(props)
            else:
                properties.extend(props)
        return properties

    def add_property(self, name, *args, **kwargs):
        prop = PROPERTY_CLASSES.get(name)(*args, **kwargs)
        self._properties.append((None, prop))
        return prop

    def remove_property(self, prop):
        for _group, props in list(self._properties):
            if isinstance(props, list):
                if prop in props:
                    props.remove(prop)
                    return

            elif prop is props:
                self._properties.remove((None, props))
                return

        raise ValueError('prop not found in vcard')

    def copy(self):
        properties = []
        for group_name, props in self._properties:
            if group_name is None:
                properties.append((None, props.copy()))
            else:
                group_properties = [prop.copy() for prop in props]
                properties.append((group_name, group_properties))
        return self.__class__(properties=properties)


class VCard4(BaseModule):

    _depends = {
        'publish': 'PubSub',
        'request_items': 'PubSub',
    }

    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_vcard(self, jid=None):
        task = yield

        items = yield self.request_items(Namespace.VCARD4_PUBSUB,
                                         jid=jid,
                                         max_items=1)

        raise_if_error(items)

        if not items:
            yield task.set_result(None)

        yield _get_vcard(items[0])

    @iq_request_task
    def set_vcard(self, vcard, public=False):
        task = yield

        access_model = 'open' if public else 'presence'

        options = {
            'pubsub#persist_items': 'true',
            'pubsub#access_model': access_model,
        }

        result = yield self.publish(Namespace.VCARD4_PUBSUB,
                                    vcard.to_node(),
                                    id_='current',
                                    options=options,
                                    force_node_options=True)

        yield finalize(task, result)


def _get_vcard(item):
    vcard = item.getTag('vcard', namespace=Namespace.VCARD4)
    if vcard is None:
        raise MalformedStanzaError('vcard node missing', item)

    try:
        vcard = VCard.from_node(vcard)
    except Exception as error:
        raise MalformedStanzaError('invalid vcard: %s' % error, item)

    return vcard



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; If not, see <http://www.gnu.org/licenses/>.

from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import ERR_BAD_REQUEST
from nbxmpp.protocol import ERR_FEATURE_NOT_IMPLEMENTED
from nbxmpp.protocol import Error as ErrorStanza
from nbxmpp.protocol import Iq
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import IBBData
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task
from nbxmpp.util import b64decode
from nbxmpp.util import b64encode


class IBB(BaseModule):
    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='iq',
                          callback=self._process_ibb,
                          ns=Namespace.IBB,
                          priority=20),
        ]

    def _process_ibb(self, _client, stanza, properties):
        if properties.type.is_set:
            open_ = stanza.getTag('open', namespace=Namespace.IBB)
            if open_ is not None:
                properties.ibb = self._parse_open(stanza, open_)
                return

            close = stanza.getTag('close', namespace=Namespace.IBB)
            if close is not None:
                properties.ibb = self._parse_close(stanza, close)
                return

            data = stanza.getTag('data', namespace=Namespace.IBB)
            if data is not None:
                properties.ibb = self._parse_data(stanza, data)
                return

    def _parse_open(self, stanza, open_):
        attrs = open_.getAttrs()
        try:
            block_size = int(attrs.get('block-size'))
        except Exception as error:
            self._log.warning(error)
            self._log.warning(stanza)
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        if block_size > 65535:
            self._log.warning('Invalid block-size')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        sid = attrs.get('sid')
        if not sid:
            self._log.warning('Invalid sid')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        type_ = attrs.get('stanza')
        if type_ == 'message':
            self._client.send_stanza(ErrorStanza(stanza,
                                                 ERR_FEATURE_NOT_IMPLEMENTED))
            raise NodeProcessed

        return IBBData(type='open', block_size=block_size, sid=sid)

    def _parse_close(self, stanza, close):
        sid = close.getAttrs().get('sid')
        if sid is None:
            self._log.warning('Invalid sid')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed
        return IBBData(type='close', sid=sid)

    def _parse_data(self, stanza, data):
        attrs = data.getAttrs()

        sid = attrs.get('sid')
        if sid is None:
            self._log.warning('Invalid sid')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        try:
            seq = int(attrs.get('seq'))
        except Exception:
            self._log.exception('Invalid seq')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        try:
            decoded_data = b64decode(data.getData())
        except Exception:
            self._log.exception('Failed to decode IBB data')
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

        return IBBData(type='data', sid=sid, seq=seq, data=decoded_data)

    def send_reply(self, stanza, error=None):
        if error is None:
            reply = stanza.buildReply('result')
            reply.getChildren().clear()
        else:
            reply = ErrorStanza(stanza, error)
        self._client.send_stanza(reply)

    @iq_request_task
    def send_open(self, jid, sid, block_size):
        _task = yield

        response = yield _make_ibb_open(jid, sid, block_size)
        yield process_response(response)

    @iq_request_task
    def send_close(self, jid, sid):
        _task = yield

        response = yield _make_ibb_close(jid, sid)
        yield process_response(response)

    @iq_request_task
    def send_data(self, jid, sid, seq, data):
        _task = yield

        response = yield _make_ibb_data(jid, sid, seq, data)
        yield process_response(response)


def _make_ibb_open(jid, sid, block_size):
    iq = Iq('set', to=jid)
    iq.addChild('open',
                {'block-size': block_size, 'sid': sid, 'stanza': 'iq'},
                namespace=Namespace.IBB)
    return iq


def _make_ibb_close(jid, sid):
    iq = Iq('set', to=jid)
    iq.addChild('close', {'sid': sid}, namespace=Namespace.IBB)
    return iq


def _make_ibb_data(jid, sid, seq, data):
    iq = Iq('set', to=jid)
    ibb_data = iq.addChild('data',
                           {'sid': sid, 'seq': seq},
                           namespace=Namespace.IBB)
    ibb_data.setData(b64encode(data))
    return iq



from nbxmpp.errors import MalformedStanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.protocol import Node
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import MDSData
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

MDS_OPTIONS = {
    'pubsub#persist_items': 'true',
    'pubsub#max_items': 'max',
    'pubsub#send_last_published_item': 'never',
    'pubsub#access_model': 'whitelist',
}


class MDS(BaseModule):
    """
    XEP-0490
    """
    _depends = {
        'publish': 'PubSub'
    }

    def __init__(self, client):
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(name='message',
                          callback=self._process_pubsub_mds,
                          ns=Namespace.PUBSUB_EVENT,
                          priority=16),
        ]

    def _process_pubsub_mds(self, _client, stanza, properties):
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.MDS:
            return

        item = properties.pubsub_event.item
        if item is None:
            return

        try:
            data = self._parse_item(item)
        except MalformedStanzaError as error:
            self._log.warning(error)
            self._log.warning(stanza)
            raise NodeProcessed

        pubsub_event = properties.pubsub_event._replace(data=data)
        self._log.info('Received MDS: %s', data)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def set_mds(self, data: MDSData, by: JID):
        task = yield

        item = Node('displayed', {'xmlns': Namespace.MDS})
        sid = item.addChild('stanza-id', namespace=Namespace.SID)
        sid.setAttr('id', data.stanza_id)
        sid.setAttr('by', str(by))

        result = yield self.publish(Namespace.MDS,
                                    item,
                                    id_=str(data.jid),
                                    options=MDS_OPTIONS,
                                    force_node_options=True)

        yield finalize(task, result)

    @iq_request_task
    def get_mds(self):
        task = yield

        items = yield self.request_items(Namespace.MDS)

        raise_if_error(items)

        if not items:
            yield task.set_result(None)

        data: list[MDSData] = []
        for item in items:
            try:
                data.append(self._parse_item(item))
            except MalformedStanzaError as error:
                self._log.warning(error)
                self._log.warning(item)

        yield data

    def _parse_item(self, item: Node) -> MDSData:

        item_id = item.getAttr('id')
        try:
            jid = JID.from_string(item_id)
        except Exception as e:
            raise MalformedStanzaError(
                f'MDS item ID "{item_id}" is not a valid JID: {e}',
                item)

        displayed = item.getTag('displayed', namespace=Namespace.MDS)
        if not displayed:
            raise MalformedStanzaError(
                'Bad MDS event (no displayed tag)', item)

        sid = displayed.getTag('stanza-id', namespace=Namespace.SID)
        if sid is None:
            raise MalformedStanzaError(
                'Bad MDS event (no stanza-id tag)', item)

        stanza_id = sid.getAttr('id')
        if not stanza_id:
            raise MalformedStanzaError(
                'Bad MDS event (no stanza-id ID)', item)

        by = sid.getAttr('by')
        try:
            stanza_id_by = JID.from_string(by)
        except Exception as e:
            raise MalformedStanzaError(
                f'MDS stanza-id by "{by}" is not a valid JID: {e}', item)

        return MDSData(
            jid=jid,
            stanza_id=stanza_id,
            stanza_id_by=stanza_id_by
        )
